#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <allegro.h>

#include "cpu.h"
#include "tipi.h"
#include "vmachine.h"
#include "audio.h"

#define SAMPLE_RATE 		44100
#define PERIOD1 				11
#define PERIOD2 				44

#define SOUND_BUFFER_LEN 	1056

#define AUD_CTRL  			0xAA
#define AUD_D0	  				0xA7
#define AUD_D1	  				0xA8
#define AUD_D2	  				0xA9

int Sound_IRQ;
static AUDIOSTREAM *Stream=NULL;
FILE *SndLog=NULL;

static double Flt_A=0.0,Flt_B=0.0;
static unsigned char Flt_Prv=0;

//////////////////////////////////////////////////////////////////////////////////

static void Filter(unsigned char *Buf,unsigned long Len);

void Audio_Process(unsigned char *Buffer)
{
	unsigned long Aud_Data;
	int Volume,Re_Circ,Noise,Enabled,IntEna,Period,Pnt,Cnt,RndBit,Pos;
	
	Aud_Data=(VDCWrite[AUD_D2] | (VDCWrite[AUD_D1] << 8) | (VDCWrite[AUD_D0] << 16));
	IntEna=VDCWrite[0xA0] & 0x04;
	Pnt=Cnt=0;
	Noise=VDCWrite[AUD_CTRL] & 0x10;
	Enabled=VDCWrite[AUD_CTRL] & 0x80;
	RndBit=(Enabled && Noise) ? (rand() % 2) : 0;
	
	while (Pnt < SOUND_BUFFER_LEN)
	{
		Pos=(TweakedAudio) ? (Pnt / 3) : (MAXLINES - 1);
		Volume=AudioVector[Pos] & 0x0F;
		Enabled=AudioVector[Pos] & 0x80;
		Period=(AudioVector[Pos] & 0x20) ? PERIOD1 : PERIOD2;
		Re_Circ=AudioVector[Pos] & 0x40;
		
		Buffer[Pnt++]=(Enabled) ? ((Aud_Data & 0x01)^RndBit) * (0x10 * Volume) : 0;
		Cnt++;
		
		if (Cnt >= Period)
		{
			Cnt=0;
			Aud_Data=(Re_Circ) ? ((Aud_Data >> 1) | ((Aud_Data & 1) << 23)) : (Aud_Data >>= 1);
			RndBit=(Enabled && Noise) ? (rand() % 2) : 0;
			
			if (Enabled && IntEna && (!Sound_IRQ))
			{
				Sound_IRQ=1;
				
				Ext_IRQ();
			}
		}
	}
	
	if (App_Data.Filter)
		Filter(Buffer,SOUND_BUFFER_LEN);
}

void UpDate_Audio(void)
{
	unsigned char *P;
	
	if (App_Data.Sound_En)
	{
		P=(unsigned char *)get_audio_stream_buffer(Stream);
		
		if (P)
		{
			Audio_Process(P);
			
			if (SndLog)
				fwrite(P,1,SOUND_BUFFER_LEN,SndLog);
			
			free_audio_stream_buffer(Stream);
		}
	}
}

void Init_Audio(void)
{
	int I;
	
	Sound_IRQ=0;
	
	if ((App_Data.Sound_En) || (App_Data.Voice))
	{
		printf("Starting sound system...\n");
		
		I=install_sound(DIGI_AUTODETECT,MIDI_NONE,NULL);
		
		if (I != 0)
		{
			printf("  ERROR: could not initialize sound!\n   %s\n",allegro_error);
			printf("  Sound system disabled\n");
			
			App_Data.Sound_En=0;
			App_Data.Voice=0;
		}
		else
		{
			if (digi_driver->name && (strlen(digi_driver->name)>0))
			{
				printf("  Sound system running !\n");
				printf("  Sound driver [%s] detected\n",digi_driver->name);
				
				set_volume(255,255);
				Init_Sound_Stream();
			}
			else
			{
				printf("  ERROR: could not initialize sound card\n");
				printf("  Sound system disabled\n");
				
				App_Data.Sound_En=0;
				App_Data.Voice=0;
			}
		}
	}
	
	SndLog=NULL;
}

void Init_Sound_Stream(void)
{
	int Vol;
	
	if (App_Data.Sound_En)
	{
		if (App_Data.Filter)
			Vol=(255 * App_Data.SVolume) / 100;
		else
			Vol=(255 * App_Data.SVolume) / 200;
		
		Stream=play_audio_stream(SOUND_BUFFER_LEN,8,0,SAMPLE_RATE,Vol,128);
		
		if (!Stream)
		{
			printf("Error creating audio stream!\n");
			
			App_Data.Sound_En=0;
		}
		
		Flt_A=Flt_B=0.0;
		Flt_Prv=0;
	}
}

void Mute_Audio(void)
{
	if (App_Data.Sound_En && Stream)
	{
		stop_audio_stream(Stream);
		Stream=NULL;
	}
}

void Close_Audio(void)
{
	if (App_Data.Sound_En && Stream)
		stop_audio_stream(Stream);
	
	if (SndLog)
		fclose(SndLog);
	
	App_Data.Sound_En=0;
}

static void Filter(unsigned char *Buf,unsigned long Len)
{
	static unsigned char buf[SOUND_BUFFER_LEN];
	int T;
	register unsigned long I;
	
	if (Len > SOUND_BUFFER_LEN)
		return;
	
	memcpy(buf,Buf,Len);
	
	for (I=0; I < Len; I++)
	{
		T=(I == 0) ? (buf[0] - Flt_Prv) : (buf[I] - buf[I - 1]);
		
		if (T)
			Flt_B=(double)T;
		
		Flt_A+=Flt_B / 4.0 - Flt_A / 80.0;
		Flt_B-=Flt_B / 4.0;
		
		if ((Flt_A > 255.0) || (Flt_A < -255.0))
			Flt_A=0.0;
		
		Buf[I]=(unsigned char)((Flt_A + 255.0) / 2.0);
	}
	
	Flt_Prv=buf[Len - 1];
}
