#include <stdio.h>

#include "tipi.h"
#include "vmachine.h"
#include "keyboard.h"
#include "voice.h"
#include "vdc.h"
#include "vpp.h"
#include "cpu.h"

Byte Acc;
Indirizzo Pc;
long Clk;

Byte ITimer;
Byte Reg_Pnt;
Byte Timer_On;
Byte Count_On;

Byte T_Flag;
Byte Psw;
Byte Sp;
Byte P1,P2;

Byte xIrq_pEnd,tIrq_pEnd;

static Indirizzo LastPc;
static Indirizzo A11;
static Indirizzo A11FF;
static Byte Bs,F0,F1,Ac,Cy,xIrq_En,tIrq_En,Irq_Ex;
static int Master_Count;

#define Push(d) {IntRam[Sp++] = (d); if (Sp > 23) Sp = 8;}
#define Pull() (Sp--, (Sp < 8) ? (Sp=23) : 0, IntRam[Sp])
#define Make_Psw() {Psw = (Cy << 7) | Ac | F0 | Bs | 0x08; Psw = Psw | ((Sp - 8) >> 1);}
#define Illegal(o) {}
#define Undef(i) {printf("** unimplemented instruction %x, %x**\n",i,Pc);}


void Init_Cpu(void)
{
	Pc=0;
	Sp=8;
	Bs=0;
	P1=P2=0xFF;
	Ac=Cy=F0=0;
	A11=A11FF=0;
	Timer_On=0;
	Count_On=0;
	Reg_Pnt=0;
	tIrq_En=xIrq_En=Irq_Ex=xIrq_pEnd=tIrq_pEnd=0;
}

void Ext_IRQ(void)
{
	Int_Clk=5;
	
	if (xIrq_En && !Irq_Ex)
	{
		Irq_Ex=1;
		xIrq_pEnd=0;
		Clk+=2;
		
		Make_Psw();		
		Push(Pc & 0xFF);
		Push(((Pc & 0xF00) >> 8) | (Psw & 0xF0));
		
		Pc=0x03;
		A11FF=A11;
		A11=0;
	}
	
	if (pEndIrq && (!xIrq_En))		
		xIrq_pEnd=1;
}

void Tim_IRQ(void)
{
	if (tIrq_En && !Irq_Ex)
	{
		Irq_Ex=2;
		tIrq_pEnd=0;
		Clk+=2;
		
		Make_Psw();
		Push(Pc & 0xFF);
		Push(((Pc & 0xF00) >> 8) | (Psw & 0xF0));
		
		Pc=0x07;
		A11FF=A11;
		A11=0;
	}
	
	if (pEndIrq && (!tIrq_En))
		tIrq_pEnd=1;
}

void Make_Psw_Debug(void)
{
	Make_Psw();
}

void Cpu_Exec(void)
{
	register Byte Op;
	Indirizzo Adr;
	Byte Dat;
	int Temp;
	
	for (;;)
	{
		Clk=0;
		LastPc=Pc;
		Op=Rom[Pc++];
		
		switch (Op)
		{
			case 0x00:	//Nop
				Clk++;
				break;
			
			case 0x01:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x02:	//OUTL BUS,A
				Clk+=2;
				Undef(0x02);
				break;
			
			case 0x03:	//ADD A,#data
				Clk+=2;
				Cy=Ac=0;
				Dat=Rom[Pc++];
				
				if (((Acc & 0x0f) + (Dat & 0x0f)) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x04:	//JMP
				Pc=Rom[Pc] | A11;
				Clk+=2;
				break;
			
			case 0x05:	//EN I
				xIrq_En=1;
				Clk++;
				break;
			
			case 0x06:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x07:	//DEC A
				Acc--;
				Clk++;
				break;
			
			case 0x08:	//INS A,BUS
				Clk+=2;
				Acc=In_Bus();
				break;
			
			case 0x09:	//IN A,Pp
				Acc=P1;
				Clk+=2;
				break;
			
			case 0x0A:	//IN A,Pp
				Acc=Read_P2();
				Clk+=2;
				break;
			
			case 0x0B:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x0C:	//MOVD A,P4
				Clk+=2;
				Acc=Read_PB(0);
				break;
			
			case 0x0D:	//MOVD A,P5
				Clk+=2;
				Acc=Read_PB(1);
				break;
			
			case 0x0E:	//MOVD A,P6
				Clk+=2;
				Acc=Read_PB(2);
				break;
			
			case 0x0F:	//MOVD A,P7
				Clk+=2;
				Acc=Read_PB(3);
				break;
			
			case 0x10:	//INC @Ri
				IntRam[IntRam[Reg_Pnt] & 0x3F]++;
				Clk++;
				break;
			
			case 0x11:	//INC @Ri
				IntRam[IntRam[Reg_Pnt + 1] & 0x3F]++;
				Clk++;
				break;
			
			case 0x12:	//JBb address
				Clk+=2;
				Dat=Rom[Pc];
			
				if (Acc & 0x01)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0x13:	//ADDC A,#data
				Clk+=2;
				Dat=Rom[Pc++];
				Ac=0;
				
				if (((Acc & 0x0f) + (Dat & 0x0f) + Cy) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat + Cy;
				Cy=0;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x14:	//CALL
				Make_Psw();
				
				Adr=Rom[Pc] | A11;
				Pc++;
				Clk+=2;
			
				Push(Pc & 0xFF);
				Push(((Pc & 0xF00) >> 8) | (Psw & 0xF0));
				Pc=Adr;
				break;
			
			case 0x15:	//DIS I
				xIrq_En=0;
				Clk++;
				break;
			
			case 0x16:	//JTF
				Clk+=2;
				Dat=Rom[Pc];
				
				if (T_Flag)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				
				T_Flag=0;
				break;
				
			case 0x17:	//INC A
				Acc++;
				Clk++;
				break;
			
			case 0x18:	//INC Rr
				IntRam[Reg_Pnt]++;
				Clk++;
				break;
			
			case 0x19:	//INC Rr
				IntRam[Reg_Pnt + 1]++;
				Clk++;
				break;
			
			case 0x1A:	//INC Rr
				IntRam[Reg_Pnt + 2]++;
				Clk++;
				break;
			
			case 0x1B:	//INC Rr
				IntRam[Reg_Pnt + 3]++;
				Clk++;
				break;
			
			case 0x1C:	//INC Rr
				IntRam[Reg_Pnt + 4]++;
				Clk++;
				break;
			
			case 0x1D:	//INC Rr
				IntRam[Reg_Pnt + 5]++;
				Clk++;
				break;
			
			case 0x1E:	//INC Rr
				IntRam[Reg_Pnt + 6]++;
				Clk++;
				break;
			
			case 0x1F:	//INC Rr
				IntRam[Reg_Pnt + 7]++;
				Clk++;
				break;
			
			case 0x20:	//XCH A,@Ri
				Clk++;
				Dat=Acc;
				Acc=IntRam[IntRam[Reg_Pnt] & 0x3F];
				IntRam[IntRam[Reg_Pnt] & 0x3F]=Dat;
				break;
			
			case 0x21:	//XCH A,@Ri
				Clk++;
				Dat=Acc;
				Acc=IntRam[IntRam[Reg_Pnt + 1] & 0x3F];
				IntRam[IntRam[Reg_Pnt + 1] & 0x3F]=Dat;
				break;
			
			case 0x22:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x23:	//MOV a,#data
				Clk+=2;
				Acc=Rom[Pc++];
				break;
			
			case 0x24:	//JMP
				Pc=Rom[Pc] | 0x100 | A11;
				Clk+=2;
				break;
			
			case 0x25:	//EN TCNTI
				tIrq_En=1;
				Clk++;
				break;
			
			case 0x26:	//JNT0
				Clk+=2;
				Dat=Rom[Pc];
			
				if (!Get_Voice_Status())
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0x27:	//CLR A
				Clk++;
				Acc=0;
				break;
			
			case 0x28:	//XCH A,Rr
				Dat=Acc;
				Acc=IntRam[Reg_Pnt];
				IntRam[Reg_Pnt]=Dat;
				Clk++;
				break;
			
			case 0x29:	//XCH A,Rr
				Dat=Acc;
				Acc=IntRam[Reg_Pnt + 1];
				IntRam[Reg_Pnt + 1]=Dat;
				Clk++;
				break;
			
			case 0x2A:	//XCH A,Rr
				Dat=Acc;
				Acc=IntRam[Reg_Pnt + 2];
				IntRam[Reg_Pnt + 2]=Dat;
				Clk++;
				break;
			
			case 0x2B:	//XCH A,Rr
				Dat=Acc;
				Acc=IntRam[Reg_Pnt + 3];
				IntRam[Reg_Pnt + 3]=Dat;
				Clk++;
				break;
			
			case 0x2C:	//XCH A,Rr
				Dat=Acc;
				Acc=IntRam[Reg_Pnt + 4];
				IntRam[Reg_Pnt + 4]=Dat;
				Clk++;
				break;
			
			case 0x2D:	//XCH A,Rr
				Dat=Acc;
				Acc=IntRam[Reg_Pnt + 5];
				IntRam[Reg_Pnt + 5]=Dat;
				Clk++;
				break;
			
			case 0x2E:	//XCH A,Rr
				Dat=Acc;
				Acc=IntRam[Reg_Pnt + 6];
				IntRam[Reg_Pnt + 6]=Dat;
				Clk++;
				break;
			
			case 0x2F:	//XCH A,Rr
				Dat=Acc;
				Acc=IntRam[Reg_Pnt +7];
				IntRam[Reg_Pnt + 7]=Dat;
				Clk++;
				break;
			
			case 0x30:	//XCHD A,@Ri
				Clk++;
				Adr=IntRam[Reg_Pnt] & 0x3F;
				Dat=Acc & 0x0F;
				Acc&=0xF0;
				Acc=Acc | (IntRam[Adr] & 0x0F);
				IntRam[Adr]&=0xF0;
				IntRam[Adr]|=Dat;
				break;
			
			case 0x31:	//XCHD A,@Ri
				Clk++;
				Adr=IntRam[Reg_Pnt + 1] & 0x3F;
				Dat=Acc & 0x0F;
				Acc&=0xF0;
				Acc=Acc | (IntRam[Adr] & 0x0F);
				IntRam[Adr]&=0xF0;
				IntRam[Adr]|=Dat;
				break;
			
			case 0x32:	//JBb address
				Clk+=2;
				Dat=Rom[Pc];
				
				if (Acc & 0x02)
					Pc=(Pc & 0xF00) | Dat;
				else
						Pc++;
				break;
				
			case 0x33:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x34:	//CALL
				Make_Psw();
			
				Adr=Rom[Pc] | 0x100 | A11;
				Pc++;
				Clk+=2;
				
				Push(Pc & 0xFF);
				Push(((Pc & 0xF00) >> 8) | (Psw & 0xF0));
			
				Pc=Adr;
				break;
			
			case 0x35:	//DIS TCNTI
				tIrq_En=0;
				tIrq_pEnd=0;
				Clk++;
				break;
			
			case 0x36:	//JT0
				Clk+=2;
				Dat=Rom[Pc];
				
				if (Get_Voice_Status())
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0x37:	//CPL A
				Acc^=0xFF;
				Clk++;
				break;
			
			case 0x38:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x39:	//OUTL P1,A
				Clk+=2;
			
				Write_P1(Acc);
			
				break;
			
			case 0x3A:	//OUTL P2,A
				Clk+=2;
				P2=Acc;
				break;
			
			case 0x3B:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x3C:	//MOVD P4,A
				Clk+=2;
			
				Write_PB(0,Acc);
				
				break;
			
			case 0x3D:	//MOVD P5,A
				Clk+=2;
			
				Write_PB(1,Acc);
				
				break;
			
			case 0x3E:	//MOVD P6,A
				Clk+=2;
			
				Write_PB(2,Acc);
				
				break;
			
			case 0x3F:	//MOVD P7,A
				Clk+=2;
			
				Write_PB(3,Acc);
				
				break;
			
			case 0x40:	//ORL A,@Ri
				Clk++;
				Acc|=IntRam[IntRam[Reg_Pnt] & 0x3F];
				break;
			
			case 0x41:	//ORL A,@Ri
				Clk++;
				Acc|=IntRam[IntRam[Reg_Pnt + 1] & 0x3F];
				break;
			
			case 0x42:	//MOV A,T
				Clk++;	
				Acc=ITimer;
				break;
			
			case 0x43:	//ORL A,#data
				Clk+=2;
				Acc|=Rom[Pc++];
				break;
			
			case 0x44:	//JMP
				Pc=Rom[Pc] | 0x200 | A11;
				Clk+=2;
				break;
			
			case 0x45:	//STRT CNT
				Count_On=1;
				Clk++;
				break;
			
			case 0x46:	//JNT1
				Clk+=2;
				Dat=Rom[Pc];
				
				if (!Read_T1())
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0x47:	//SWAP A
				Clk++;
				Dat=(Acc & 0xF0) >> 4;
				Acc<<=4;
				Acc|=Dat;
				break;
			
			case 0x48:	//ORL A,Rr
				Clk++;
				Acc|=IntRam[Reg_Pnt];
				break;
			
			case 0x49:	//ORL A,Rr
				Clk++;
				Acc|=IntRam[Reg_Pnt + 1];
				break;
			
			case 0x4A:	//ORL A,Rr
				Clk++;
				Acc|=IntRam[Reg_Pnt + 2];
				break;
			
			case 0x4B:	//ORL A,Rr
				Clk++;
				Acc|=IntRam[Reg_Pnt + 3];
				break;
			
			case 0x4C:	//ORL A,Rr
				Clk++;
				Acc|=IntRam[Reg_Pnt + 4];
				break;
			
			case 0x4D:	//ORL A,Rr
				Clk++;
				Acc|=IntRam[Reg_Pnt + 5];
				break;
			
			case 0x4E:	//ORL A,Rr
				Clk++;
				Acc|=IntRam[Reg_Pnt + 6];
				break;
			
			case 0x4F:	//ORL A,Rr
				Clk++;
				Acc|=IntRam[Reg_Pnt + 7];
				break;
			
			case 0x50:	//ANL A,@Ri
				Acc&=IntRam[IntRam[Reg_Pnt] & 0x3F];
				Clk++;
				break;
			
			case 0x51:	//ANL A,@Ri
				Acc&=IntRam[IntRam[Reg_Pnt + 1] & 0x3F];
				Clk++;
				break;
			
			case 0x52:	//JBb address
				Clk+=2;
				Dat=Rom[Pc];
				
				if (Acc & 0x04)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0x53:	//ANL A,#data
				Clk+=2;
				Acc&=Rom[Pc++];
				break;
			
			case 0x54:	//CALL
				Make_Psw();
			
				Adr=Rom[Pc] | 0x200 | A11;
				Pc++;
				Clk+=2;
				
				Push(Pc & 0xFF);
				Push(((Pc & 0xF00) >> 8) | (Psw & 0xF0));
			
				Pc=Adr;
				break;
			
			case 0x55:	//STRT T
				Timer_On=1;
				Clk++;
				break;
			
			case 0x56:	//JT1
				Clk+=2;
				Dat=Rom[Pc];
			
				if (Read_T1())
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0x57:	//DA A
				Clk++;
				
				if (((Acc & 0x0F) > 0x09) || Ac)
					Acc+=6;
				
				Dat=(Acc & 0xF0) >> 4;
				
				if ((Dat > 9) || Cy)
				{
					Dat+=6;
					Cy=1;
				}
				else
					Cy=0;
				
				Acc=(Acc & 0x0F) | (Dat << 4);
				break;
				
			case 0x58:	//ANL A,Rr
				Clk++;
				Acc&=IntRam[Reg_Pnt];
				break;
			
			case 0x59:	//ANL A,Rr
				Clk++;
				Acc&=IntRam[Reg_Pnt + 1];
				break;
			
			case 0x5A:	//ANL A,Rr
				Clk++;
				Acc&=IntRam[Reg_Pnt + 2];
				break;
			
			case 0x5B:	//ANL A,Rr
				Clk++;
				Acc&=IntRam[Reg_Pnt + 3];
				break;
			
			case 0x5C:	//ANL A,Rr
				Clk++;
				Acc&=IntRam[Reg_Pnt + 4];
				break;
			
			case 0x5D:	//ANL A,Rr
				Clk++;
				Acc&=IntRam[Reg_Pnt + 5];
				break;
			
			case 0x5E:	//ANL A,Rr
				Clk++;
				Acc&=IntRam[Reg_Pnt + 6];
				break;
			
			case 0x5F:	//ANL A,Rr
				Clk++;
				Acc&=IntRam[Reg_Pnt + 7];
				break;
			
			case 0x60:	//ADD A,@Ri
				Clk++;
				Cy=Ac=0;
				Dat=IntRam[IntRam[Reg_Pnt] & 0x3F];
				
				if (((Acc & 0x0f) + (Dat & 0x0f)) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x61:	//ADD A,@Ri
				Clk++;
				Cy=Ac=0;
				Dat=IntRam[IntRam[Reg_Pnt + 1] & 0x3F];
				
				if (((Acc & 0x0f) + (Dat & 0x0f)) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x62:	//MOV T,A
				Clk++;
				ITimer=Acc;
				break;
			
			case 0x63:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x64:	//JMP
				Pc=Rom[Pc] | 0x300 | A11;
				Clk+=2;
				break;
			
			case 0x65:	//STOP TCNT
				Clk++;
				Count_On=Timer_On=0;
				break;
			
			case 0x66:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x67:	//RRC A
				Dat=Cy;
				Cy=Acc & 0x01;
				Acc>>=1;
				
				if (Dat)
					Acc|=0x80;
				else
					Acc&=0x7F;
				
				Clk++;
				break;
				
			case 0x68:	//ADD A,Rr
				Clk++;
				Cy=Ac=0;
				Dat=IntRam[Reg_Pnt];
				
				if (((Acc & 0x0f) + (Dat & 0x0f)) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x69:	//ADD A,Rr
				Clk++;
				Cy=Ac=0;
				Dat=IntRam[Reg_Pnt + 1];
				
				if (((Acc & 0x0f) + (Dat & 0x0f)) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x6A:	//ADD A,Rr
				Clk++;
				Cy=Ac=0;
				Dat=IntRam[Reg_Pnt + 2];
				
				if (((Acc & 0x0f) + (Dat & 0x0f)) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x6B:	//ADD A,Rr
				Clk++;
				Cy=Ac=0;
				Dat=IntRam[Reg_Pnt + 3];
				
				if (((Acc & 0x0f) + (Dat & 0x0f)) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x6C:	//ADD A,Rr
				Clk++;
				Cy=Ac=0;
				Dat=IntRam[Reg_Pnt + 4];
				
				if (((Acc & 0x0f) + (Dat & 0x0f)) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x6D:	//ADD A,Rr
				Clk++;
				Cy=Ac=0;
				Dat=IntRam[Reg_Pnt + 5];
				
				if (((Acc & 0x0f) + (Dat & 0x0f)) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x6E:	//ADD A,Rr
				Clk++;
				Cy=Ac=0;
				Dat=IntRam[Reg_Pnt + 6];
				
				if (((Acc & 0x0f) + (Dat & 0x0f)) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x6F:	//ADD A,Rr
				Clk++;
				Cy=Ac=0;
				Dat=IntRam[Reg_Pnt + 7];
				
				if (((Acc & 0x0f) + (Dat & 0x0f)) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x70:	//ADDC A,@Ri
				Clk++;
				Ac=0;
				Dat=IntRam[IntRam[Reg_Pnt] & 0x3F];
				
				if (((Acc & 0x0f) + (Dat & 0x0f) + Cy) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat + Cy;
				Cy=0;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x71:	//ADDC A,@Ri
				Clk++;
				Ac=0;
				Dat=IntRam[IntRam[Reg_Pnt + 1] & 0x3F];
				
				if (((Acc & 0x0f) + (Dat & 0x0f) + Cy) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat + Cy;
				Cy=0;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x72:	//JBb address
				Clk+=2;
				Dat=Rom[Pc];
				
				if (Acc & 0x08)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0x73:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x74:	//CALL
				Make_Psw();
			
				Adr=Rom[Pc] | 0x300 | A11;
				Pc++;
				Clk+=2;
				
				Push(Pc & 0xFF);
				Push(((Pc & 0xF00) >> 8) | (Psw & 0xF0));
				Pc=Adr;
				break;
			
			case 0x75:	//EN CLK
				Clk++;
				Undef(Op);
				break;
			
			case 0x76:	//JF1 address
				Clk+=2;
				Dat=Rom[Pc];
			
				if (F1)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0x77:	//RR A
				Clk++;
				Dat=Acc & 0x01;
				Acc>>=1;
				
				if (Dat)
					Acc|=0x80;
				else
					Acc&=0x7f;
				break;
				
			case 0x78:	//ADDC A,Rr
				Clk++;
				Ac=0;
				Dat=IntRam[Reg_Pnt];
				
				if (((Acc & 0x0f) + (Dat & 0x0f) + Cy) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat + Cy;
				Cy=0;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x79:	//ADDC A,Rr
				Clk++;
				Ac=0;
				Dat=IntRam[Reg_Pnt + 1];
				
				if (((Acc & 0x0f) + (Dat & 0x0f) + Cy) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat + Cy;
				Cy=0;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x7A:	//ADDC A,Rr
				Clk++;
				Ac=0;
				Dat=IntRam[Reg_Pnt + 2];
				
				if (((Acc & 0x0f) + (Dat & 0x0f) + Cy) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat + Cy;
				Cy=0;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x7B:	//ADDC A,Rr
				Clk++;
				Ac=0;
				Dat=IntRam[Reg_Pnt + 3];
				
				if (((Acc & 0x0f) + (Dat & 0x0f) + Cy) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat + Cy;
				Cy=0;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x7C:	//ADDC A,Rr
				Clk++;
				Ac=0;
				Dat=IntRam[Reg_Pnt + 4];
				
				if (((Acc & 0x0f) + (Dat & 0x0f) + Cy) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat + Cy;
				Cy=0;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x7D:	//ADDC A,Rr
				Clk++;
				Ac=0;
				Dat=IntRam[Reg_Pnt + 5];
				
				if (((Acc & 0x0f) + (Dat & 0x0f) + Cy) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat + Cy;
				Cy=0;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x7E:	//ADDC A,Rr
				Clk++;
				Ac=0;
				Dat=IntRam[Reg_Pnt + 6];
				
				if (((Acc & 0x0f) + (Dat & 0x0f) + Cy) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat + Cy;
				Cy=0;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x7F:	//ADDC A,Rr
				Clk++;
				Ac=0;
				Dat=IntRam[Reg_Pnt + 7];
				
				if (((Acc & 0x0f) + (Dat & 0x0f) + Cy) > 0x0f)
					Ac=0x40;
				
				Temp=Acc + Dat + Cy;
				Cy=0;
				
				if (Temp > 0xFF)
					Cy=1;
				
				Acc=(Temp & 0xFF);
				break;
				
			case 0x80:	//MOVX  A,@Ri
				Acc=Ext_Read(IntRam[Reg_Pnt]);
				Clk+=2;
				break;
			
			case 0x81:	//MOVX  A,@Ri
				Acc=Ext_Read(IntRam[Reg_Pnt + 1]);
				Clk+=2;
				break;
			
			case 0x82:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x83:	//RET
				Pc=((Pull() & 0x0F) << 8);
				Pc|=Pull();
				break;
			
			case 0x84:	//JMP
				Pc=Rom[Pc] | 0x400 | A11;
				Clk+=2;
				break;
			
			case 0x85:	//CLR F0
				Clk++;
				F0=0;
				break;
			
			case 0x86:	//JNI address
				Clk+=2;
				Dat=Rom[Pc];
			
				if (Int_Clk > 0)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0x87:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x88:	//BUS,#data
				Clk+=2;
				Undef(Op);
				break;
			
			case 0x89:	//ORL Pp,#data
				Write_P1(P1 | Rom[Pc++]);
				Clk+=2;
				break;
			
			case 0x8A:	//ORL Pp,#data
				P2|=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0x8B:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x8C:	//ORLD P4,A
				Write_PB(0,Read_PB(0) | Acc);
				Clk+=2;
				break;
			
			case 0x8D:	//ORLD P5,A
				Write_PB(1,Read_PB(1) | Acc);
				Clk+=2;
				break;
			
			case 0x8E:	//ORLD P6,A
				Write_PB(2,Read_PB(2) | Acc);
				Clk+=2;
				break;
			
			case 0x8F:	//ORLD P7,A
				Write_PB(3,Read_PB(3) | Acc);
				Clk+=2;
				break;
			
			case 0x90:	//MOVX @Ri,A
				Ext_Write(Acc,IntRam[Reg_Pnt]);
				Clk+=2;
				break;
			
			case 0x91:	//MOVX @Ri,A
				Ext_Write(Acc,IntRam[Reg_Pnt + 1]);
				Clk+=2;
				break;
			
			case 0x92:	//JBb address
				Clk+=2;
				Dat=Rom[Pc];
				
				if (Acc & 0x10)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0x93:	//RETR
				Clk+=2;
				Dat=Pull();
				Pc=(Dat & 0x0F) << 8;
				Cy=(Dat & 0x80) >> 7;
				Ac=Dat & 0x40;
				F0=Dat & 0x20;
				Bs=Dat & 0x10;
			
				if (Bs)
					Reg_Pnt=24;
				else
					Reg_Pnt=0;
				
				Pc|=Pull();
				Irq_Ex=0;
				A11=A11FF;
				break;
				
			case 0x94:	//CALL
				Make_Psw();
			
				Adr=Rom[Pc] | 0x400 | A11;
				Pc++;
				Clk+=2;
				
				Push(Pc & 0xFF);
				Push(((Pc & 0xF00) >> 8) | (Psw & 0xF0));
			
				Pc=Adr;
				break;
			
			case 0x95:	//CPL F0
				F0^=0x20;
				Clk++;
			
			case 0x96:	//JNZ address
				Clk+=2;
				Dat=Rom[Pc];
				
				if (Acc != 0)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0x97:	//CLR C
				Cy=0;
				Clk++;
				break;
			
			case 0x98:	//ANL BUS,#data
				Clk+=2;
				Undef(Op);
				break;
			
			case 0x99:	//ANL Pp,#data
				Write_P1(P1 & Rom[Pc++]);
				Clk+=2;
				break;
			
			case 0x9A:	//ANL Pp,#data
				P2&=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0x9B:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0x9C:	//ANLD P4,A
				Write_PB(0,Read_PB(0) & Acc);
				Clk+=2;
				break;
			
			case 0x9D:	//ANLD P4,A
				Write_PB(1,Read_PB(1) & Acc);
				Clk+=2;
				break;
			
			case 0x9E:	//ANLD P4,A
				Write_PB(2,Read_PB(2) & Acc);
				Clk+=2;
				break;
			
			case 0x9F:	//ANLD P4,A
				Write_PB(3,Read_PB(3) & Acc);
				Clk+=2;
				break;
			
			case 0xA0:	//MOV @Ri,A
				IntRam[IntRam[Reg_Pnt] & 0x3F]=Acc;
				Clk++;
				break;
			
			case 0xA1: 	//MOV @Ri,A
				IntRam[IntRam[Reg_Pnt + 1] & 0x3F]=Acc;
				Clk++;
				break;
			
			case 0xA2:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xA3: 	//MOVP A,@A
				Acc=Rom[(Pc & 0xF00) | Acc];
				Clk+=2;
				break;
			
			case 0xA4:	//JMP
				Pc=Rom[Pc] | 0x500 | A11;
				Clk+=2;
				break;
			
			case 0xA5:	//CLR F1
				Clk++;
				F1=0;
				break;
			
			case 0xA6:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xA7:	//CPL C
				Cy^=0x01;
				Clk++;
				break;
			
			case 0xA8:	//MOV Rr,A
				IntRam[Reg_Pnt]=Acc;
				break;
			
			case 0xA9:	//MOV Rr,A
				IntRam[Reg_Pnt + 1]=Acc;
				break;
			
			case 0xAA:	//MOV Rr,A
				IntRam[Reg_Pnt + 2]=Acc;
				break;
			
			case 0xAB:	//MOV Rr,A
				IntRam[Reg_Pnt + 3]=Acc;
				break;
			
			case 0xAC:	//MOV Rr,A
				IntRam[Reg_Pnt + 4]=Acc;
				break;
			
			case 0xAD:	//MOV Rr,A
				IntRam[Reg_Pnt + 5]=Acc;			
				break;
			
			case 0xAE:	//MOV Rr,A
				IntRam[Reg_Pnt + 6]=Acc;
				break;
			
			case 0xAF:	//MOV Rr,A
				IntRam[Reg_Pnt + 7]=Acc;
				break;
			
			case 0xB0:	//MOV @Ri,#data
				IntRam[IntRam[Reg_Pnt] & 0x3F]=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0xB1:	//MOV @Ri,#data
				IntRam[IntRam[Reg_Pnt + 1] & 0x3F]=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0xB2:	//JBb address
				Clk+=2;
				Dat=Rom[Pc];
				
				if (Acc & 0x20)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0xB3:	//JMPP @A
				Adr=(Pc & 0xF00) | Acc;
				Pc=(Pc & 0xF00) | Rom[Adr];
				Clk+=2;
				break;
			
			case 0xB4:	//CALL
				Make_Psw();
			
				Adr=Rom[Pc] | 0x500 | A11;
				Pc++;
				Clk+=2;
				
				Push(Pc & 0xFF);
				Push(((Pc & 0xF00) >> 8) | (Psw & 0xF0));
			
				Pc=Adr;
				break;
			
			case 0xB5:	//CPL F1
				F1^=0x01;
				Clk++;
				break;
			
			case 0xB6:	//JF0 address
				Clk+=2;
				Dat=Rom[Pc];
				
				if (F0)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0xB7:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xB8:	//MOV Rr,#data
				IntRam[Reg_Pnt]=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0xB9:	//MOV Rr,#data
				IntRam[Reg_Pnt + 1]=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0xBA:	//MOV Rr,#data
				IntRam[Reg_Pnt + 2]=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0xBB:	//MOV Rr,#data
				IntRam[Reg_Pnt + 3]=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0xBC:	//MOV Rr,#data
				IntRam[Reg_Pnt + 4]=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0xBD:	//MOV Rr,#data
				IntRam[Reg_Pnt + 5]=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0xBE:	//MOV Rr,#data
				IntRam[Reg_Pnt + 6]=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0xBF:	//MOV Rr,#data
				IntRam[Reg_Pnt + 7]=Rom[Pc++];
				Clk+=2;
				break;
			
			case 0xC0:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xC1:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xC2:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xC3:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xC4:	//JMP
				Pc=Rom[Pc] | 0x600 | A11;
				Clk+=2;
				break;
			
			case 0xC5:	//SEL RB0
				Bs=Reg_Pnt=0;
				Clk++;
				break;
			
			case 0xC6:	//JZ address
				Clk+=2;
				Dat=Rom[Pc];
			
				if (Acc == 0)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0xC7:	//MOV A,PSW
				Clk++;
				
				Make_Psw();
			
				Acc=Psw;
				break;
			
			case 0xC8:	//DEC Rr
				IntRam[Reg_Pnt]--;
				Clk++;
				break;
			
			case 0xC9:	//DEC Rr
				IntRam[Reg_Pnt + 1]--;
				Clk++;
				break;
			
			case 0xCA:	//DEC Rr
				IntRam[Reg_Pnt + 2]--;
				Clk++;
				break;
			
			case 0xCB:	//DEC Rr
				IntRam[Reg_Pnt + 3]--;
				Clk++;
				break;
			
			case 0xCC:	//DEC Rr
				IntRam[Reg_Pnt + 4]--;
				Clk++;
				break;
			
			case 0xCD:	//DEC Rr
				IntRam[Reg_Pnt + 5]--;
				Clk++;
				break;
			
			case 0xCE:	//DEC Rr
				IntRam[Reg_Pnt + 6]--;
				Clk++;
				break;
			
			case 0xCF:	//DEC Rr
				IntRam[Reg_Pnt + 7]--;
				Clk++;
				break;
			
			case 0xD0:	//XRL A,@Ri
				Acc^=IntRam[IntRam[Reg_Pnt] & 0x3F];
				Clk++;
				break;
			
			case 0xD1:	//XRL A,@Ri
				Acc^=IntRam[IntRam[Reg_Pnt + 1] & 0x3F];
				Clk++;
				break;
			
			case 0xD2:	//JBb address
				Clk+=2;
				Dat=Rom[Pc];
			
				if (Acc & 0x40)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0xD3:	//XRL A,#data
				Clk+=2;
				Acc^=Rom[Pc++];
				break;
			
			case 0xD4:	//CALL
				Make_Psw();
			
				Adr=Rom[Pc] | 0x600 | A11;
				Pc++;
				Clk+=2;
				
				Push(Pc & 0xFF);
				Push(((Pc & 0xF00) >> 8) | (Psw & 0xF0));
				
				Pc=Adr;
				break;
			
			case 0xD5:	//SEL RB1
				Bs=0x10;
				Reg_Pnt=24;
				Clk++;
				break;
			
			case 0xD6:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xD7:	//MOV PSW,A
				Psw=Acc;
				Clk++;
				Cy=(Psw & 0x80) >> 7;
				Ac=Psw & 0x40;
				F0=Psw & 0x20;
				Bs=Psw & 0x10;
				
				if (Bs)
					Reg_Pnt=24;
				else
					Reg_Pnt=0;
				
				Sp=(Psw & 0x07) << 1;
				Sp+=8;
				break;
				
			case 0xD8:	//XRL A,Rr
				Acc^=IntRam[Reg_Pnt];
				Clk++;
				break;
			
			case 0xD9:	//XRL A,Rr
				Acc^=IntRam[Reg_Pnt + 1];
				Clk++;
				break;
			
			case 0xDA:	//XRL A,Rr
				Acc^=IntRam[Reg_Pnt + 2];
				Clk++;
				break;
			
			case 0xDB:	//XRL A,Rr
				Acc^=IntRam[Reg_Pnt + 3];
				Clk++;
				break;
			
			case 0xDC:	//XRL A,Rr
				Acc^=IntRam[Reg_Pnt + 4];
				Clk++;
				break;
			
			case 0xDD:	//XRL A,Rr
				Acc^=IntRam[Reg_Pnt + 5];
				Clk++;
				break;
			
			case 0xDE:	//XRL A,Rr
				Acc^=IntRam[Reg_Pnt + 6];
				Clk++;
				break;
			
			case 0xDF:	//XRL A,Rr
				Acc^=IntRam[Reg_Pnt + 7];
				Clk++;
				break;
			
			case 0xE0:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xE1:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xE2:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xE3:	//MOVP3 A,@A
				Adr=0x300 | Acc;
				Acc=Rom[Adr];
				Clk+=2;
				break;
			
			case 0xE4:	//JMP
				Pc=Rom[Pc] | 0x700 | A11;
				Clk+=2;
				break;
			
			case 0xE5:	//SEL MB0
				A11=0;
				A11FF=0;
				Clk++;
				break;
			
			case 0xE6:	//JNC address
				Clk+=2;
				Dat=Rom[Pc];
				
				if (!Cy)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0xE7:	//RL A
				Clk++;
				Dat=Acc & 0x80;
				Acc<<=1;
				
				if (Dat)
					Acc|=0x01;
				else
					Acc&=0xFE;
				break;
				
			case 0xE8:	//DJNZ Rr,address
				Clk+=2;
				IntRam[Reg_Pnt]--;
				Dat=Rom[Pc];
			
				if (IntRam[Reg_Pnt] != 0)
				{
					Pc&=0xF00;
					Pc|=Dat;
				}
				else
					Pc++;
				break;
				
			case 0xE9:	//DJNZ Rr,address
				Clk+=2;
				IntRam[Reg_Pnt + 1]--;
				Dat=Rom[Pc];
			
				if (IntRam[Reg_Pnt + 1] != 0)
				{
					Pc&=0xF00;
					Pc|=Dat;
				}
				else
					Pc++;
				break;
				
			case 0xEA:	//DJNZ Rr,address
				Clk+=2;
				IntRam[Reg_Pnt + 2]--;
				Dat=Rom[Pc];
			
				if (IntRam[Reg_Pnt + 2] != 0)
				{
					Pc&=0xF00;
					Pc|=Dat;
				}
				else
					Pc++;
				break;
				
			case 0xEB:	//DJNZ Rr,address
				Clk+=2;
				IntRam[Reg_Pnt + 3]--;
				Dat=Rom[Pc];
			
				if (IntRam[Reg_Pnt + 3] != 0)
				{
					Pc&=0xF00;
					Pc|=Dat;
				}
				else
					Pc++;
				break;
				
			case 0xEC:	//DJNZ Rr,address
				Clk+=2;
				IntRam[Reg_Pnt + 4]--;
				Dat=Rom[Pc];
			
				if (IntRam[Reg_Pnt + 4] != 0)
				{
					Pc&=0xF00;
					Pc|=Dat;
				}
				else
					Pc++;
				break;
				
			case 0xED:	//DJNZ Rr,address
				Clk+=2;
				IntRam[Reg_Pnt + 5]--;
				Dat=Rom[Pc];
			
				if (IntRam[Reg_Pnt + 5] != 0)
				{
					Pc&=0xF00;
					Pc|=Dat;
				}
				else
					Pc++;
				break;
				
			case 0xEE:	//DJNZ Rr,address
				Clk+=2;
				IntRam[Reg_Pnt + 6]--;
				Dat=Rom[Pc];
			
				if (IntRam[Reg_Pnt + 6] != 0)
				{
					Pc&=0xF00;
					Pc|=Dat;
				}
				else
					Pc++;
				break;
				
			case 0xEF:	//DJNZ Rr,address
				Clk+=2;
				IntRam[Reg_Pnt + 7]--;
				Dat=Rom[Pc];
			
				if (IntRam[Reg_Pnt + 7] != 0)
				{
					Pc&=0xF00;
					Pc|=Dat;
				}
				else
					Pc++;
				break;
				
			case 0xF0:	//MOV A,@Ri
				Clk++;
				Acc=IntRam[IntRam[Reg_Pnt] & 0x3F];
				break;
			
			case 0xF1:	//MOV A,@Ri
				Clk++;
				Acc=IntRam[IntRam[Reg_Pnt + 1] & 0x3F];
				break;
			
			case 0xF2:	//JBb address
				Clk+=2;
				Dat=Rom[Pc];
			
				if (Acc & 0x80)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0xF3:	//ILL
				Illegal(Op);
				Clk++;
				break;
			
			case 0xF4:	//CALL
				Clk+=2;
				
				Make_Psw();
			
				Adr=Rom[Pc] | 0x700 | A11;
				Pc++;
				
				Push(Pc & 0xFF);
				Push(((Pc & 0xF00) >> 8) | (Psw & 0xF0));
			
				Pc=Adr;
				break;
			
			case 0xF5:	//SEL MB1
				if (Irq_Ex)
					A11FF=0x800;
				else
				{
					A11=0x800;
					A11FF=0x800;
				}
				
				Clk++;
				break;
				
			case 0xF6:	//JC address
				Clk+=2;
				Dat=Rom[Pc];
			
				if (Cy)
					Pc=(Pc & 0xF00) | Dat;
				else
					Pc++;
				break;
				
			case 0xF7:	//RLC A
				Dat=Cy;
				Cy=(Acc & 0x80) >> 7;
				Acc<<=1;
				
				if (Dat)
					Acc|=0x01;
				else
					Acc&=0xFE;
				
				Clk++;
				break;
				
			case 0xF8:	//MOV A,Rr
				Clk++;
				Acc=IntRam[Reg_Pnt];
				break;
			
			case 0xF9:	//MOV A,Rr
				Clk++;
				Acc=IntRam[Reg_Pnt + 1];
				break;
			
			case 0xFA:	//MOV A,Rr
				Clk++;
				Acc=IntRam[Reg_Pnt + 2];
				break;
			
			case 0xFB:	//MOV A,Rr
				Clk++;
				Acc=IntRam[Reg_Pnt + 3];
				break;
			
			case 0xFC:	//MOV A,Rr
				Clk++;
				Acc=IntRam[Reg_Pnt + 4];
				break;
			
			case 0xFD:	//MOV A,Rr
				Clk++;
				Acc=IntRam[Reg_Pnt + 5];
				break;
			
			case 0xFE:	//MOV A,Rr
				Clk++;
				Acc=IntRam[Reg_Pnt + 6];
				break;
			
			case 0xFF:	//MOV A,Rr
				Clk++;
				Acc=IntRam[Reg_Pnt + 7];
				break;
		}
		
		Master_Clk+=Clk;
		H_Clk+=Clk;
		Clk_Counter+=Clk;
		
		if (Int_Clk > Clk)
			Int_Clk-=Clk;
		else
			Int_Clk=0;
		
		if (xIrq_pEnd)
			Ext_IRQ();
		
		if (tIrq_pEnd)
			Tim_IRQ();
		
		if (H_Clk > LINECNT - 1)
		{
			H_Clk-=LINECNT;
			
			if (EnahIrq && (VDCWrite[0xA0] & 0x01))
				Ext_IRQ();
			
			if (Count_On && MState == 0)
			{
				ITimer++;
				
				if (ITimer == 0)
				{
					T_Flag=1;
					
					Tim_IRQ();
					
					if (!CcolFlag)
					{
						Clear_Collision();
						CcolFlag=1;
					}
					
					Draw_Region();
				}
			}
		}
		
		if (Timer_On)
		{
			Master_Count+=Clk;
			
			if (Master_Count > 31)
			{
				Master_Count-=31;
				ITimer++;
				
				if (ITimer == 0)
				{
					T_Flag=1;
					
					Tim_IRQ();
				}
			}
		}
		
		if ((MState == 0) && (Master_Clk > VBLCLK))
			Handle_Vbl();
		
		if ((MState == 1) && (Master_Clk > EvblClk))
		{
			Handle_Evbl();
			
			break;
		}
		
		if (App_Data.Debug)
			break;
	}
}
